from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, TypeAlias

from tbc_video_export.common.enums import ChromaDecoder, ExportMode, VideoSystem


@dataclass(frozen=True, slots=True)
class VideoSystemData:
    """Base VideoSystem data class.

    Holds static values for the video system.

    Active line values from (default):
    vhs_decode/tools/library/tbc/lddecodemetadata.cpp

    Aspect ratio values from (default/widescreen):
    vhs_decode/tools/ld-chroma-decoder/outputwriter.cpp
    """

    size: dict[VideoSizeType, Size]
    active_lines: dict[VideoActiveLinesType, ActiveLines]
    aspect_ratio: dict[VideoAspectRatioType, AspectRatio]
    chroma_decoder: dict[ExportMode, ChromaDecoder]
    ffmpeg_config: FFmpegConfig

    @staticmethod
    def get(system: VideoSystem) -> VideoSystemData:
        """Returns VideoSystemData for the specified video format."""
        match system:
            case VideoSystem.PAL:
                return video_system_pal

            case VideoSystem.NTSC:
                return video_system_ntsc

            case VideoSystem.PAL_M:
                return video_system_palm

    @dataclass(frozen=True, slots=True)
    class Size:
        """Width/height data container."""

        width: int
        height: int

    @dataclass(frozen=True, slots=True)
    class AspectRatio:
        """Aspect ratio container."""

        horizontal: int
        vertical: int

    @dataclass(frozen=True, slots=True)
    class ActiveLines:
        """Active lines data container."""

        first_field: int
        last_field: int
        first_frame: int
        last_frame: int
        padding: int | None

    @dataclass(frozen=True, slots=True)
    class FFmpegConfig:
        """FFmpeg config data container."""

        color_range: str
        color_space: str
        color_primaries: str
        color_trc: str
        fps: str


video_system_pal = VideoSystemData(
    size={
        "default": VideoSystemData.Size(928, 576),  # unused
        "4fsc": VideoSystemData.Size(1135, 626),
    },
    active_lines={
        "default": VideoSystemData.ActiveLines(22, 308, 44, 620, None),  # unused
        "full_vertical": VideoSystemData.ActiveLines(2, 308, 2, 620, None),
        "letterbox": VideoSystemData.ActiveLines(2, 308, 118, 548, None),
        "vbi": VideoSystemData.ActiveLines(12, 308, 12, 620, None),
    },
    aspect_ratio={
        "default": VideoSystemData.AspectRatio(259, 311),  # unused
        "widescreen": VideoSystemData.AspectRatio(865, 779),
        "letterbox": VideoSystemData.AspectRatio(16, 9),
    },
    chroma_decoder={
        ExportMode.LUMA_EXTRACTED: ChromaDecoder.MONO,
        ExportMode.CHROMA_MERGE: ChromaDecoder.PAL2D,
        ExportMode.CHROMA_COMBINED: ChromaDecoder.TRANSFORM3D,
        ExportMode.CHROMA_COMBINED_LD: ChromaDecoder.TRANSFORM3D,
    },
    ffmpeg_config=VideoSystemData.FFmpegConfig(
        "tv",
        "bt470bg",
        "bt470bg",
        "bt709",
        "pal",
    ),
)

video_system_ntsc = VideoSystemData(
    size={
        "default": VideoSystemData.Size(760, 488),  # unused
        "4fsc": VideoSystemData.Size(910, 526),
    },
    active_lines={
        "default": VideoSystemData.ActiveLines(20, 259, 40, 525, None),  # unused
        "full_vertical": VideoSystemData.ActiveLines(1, 259, 2, 525, None),
        "letterbox": VideoSystemData.ActiveLines(61, 224, 122, 448, 1),  # unsure!
        "vbi": VideoSystemData.ActiveLines(15, 259, 16, 525, 1),
    },
    aspect_ratio={
        "default": VideoSystemData.AspectRatio(352, 413),  # unused
        "widescreen": VideoSystemData.AspectRatio(25, 22),
        "letterbox": VideoSystemData.AspectRatio(16, 9),
    },
    chroma_decoder={
        ExportMode.LUMA_EXTRACTED: ChromaDecoder.MONO,
        ExportMode.CHROMA_MERGE: ChromaDecoder.NTSC2D,
        ExportMode.CHROMA_COMBINED: ChromaDecoder.NTSC3D,
        ExportMode.CHROMA_COMBINED_LD: ChromaDecoder.NTSC2D,
    },
    ffmpeg_config=VideoSystemData.FFmpegConfig(
        "tv",
        "smpte170m",
        "smpte170m",
        "bt709",
        "ntsc",
    ),
)

video_system_palm = VideoSystemData(
    size={
        "default": VideoSystemData.Size(760, 488),  # unused
        "4fsc": VideoSystemData.Size(909, 526),
    },
    active_lines={
        "default": VideoSystemData.ActiveLines(20, 259, 40, 525, None),  # unused
        "full_vertical": VideoSystemData.ActiveLines(1, 259, 2, 525, None),
        "letterbox": VideoSystemData.ActiveLines(0, 0, 0, 0, 0),  # Sample required!
        "vbi": VideoSystemData.ActiveLines(16, 259, 17, 525, 1),
    },
    aspect_ratio={
        "default": VideoSystemData.AspectRatio(352, 413),  # unused
        "widescreen": VideoSystemData.AspectRatio(25, 22),
        "letterbox": VideoSystemData.AspectRatio(16, 9),
    },
    chroma_decoder={
        ExportMode.LUMA_EXTRACTED: ChromaDecoder.MONO,
        ExportMode.CHROMA_MERGE: ChromaDecoder.PAL2D,
        ExportMode.CHROMA_COMBINED: ChromaDecoder.TRANSFORM3D,
        ExportMode.CHROMA_COMBINED_LD: ChromaDecoder.TRANSFORM3D,
    },
    ffmpeg_config=VideoSystemData.FFmpegConfig(
        "tv",
        "bt470bg",
        "bt470bg",
        "bt709",
        "ntsc",
    ),
)

VideoSizeType: TypeAlias = Literal["default", "4fsc"]
VideoActiveLinesType: TypeAlias = Literal[
    "default", "full_vertical", "letterbox", "vbi"
]
VideoAspectRatioType: TypeAlias = Literal["default", "widescreen", "letterbox"]



from __future__ import annotations

import logging
from dataclasses import dataclass

from tbc_video_export.common.enums import ProcessStatus


@dataclass
class ProcessState:
    """Process state class."""

    status = ProcessStatus.NONE
    returncode = -1
    status_index = -1

    @property
    def has_run(self) -> bool:
        """Returns True if the process has run."""
        return ProcessStatus.HAS_RUN in self.status

    @has_run.setter
    def has_run(self, has_run: bool):
        if has_run:
            logging.getLogger("console").debug("Setting HAS_RUN")

            self.status |= ProcessStatus.HAS_RUN
            self.status &= ~ProcessStatus.NONE
        else:
            logging.getLogger("console").debug("Unsetting HAS_RUN")

            self.status &= ~(ProcessStatus.HAS_RUN | ProcessStatus.NONE)

    @property
    def running(self) -> bool:
        """Returns True if the process is running."""
        return ProcessStatus.RUNNING in self.status

    @running.setter
    def running(self, running: bool):
        """Set running state."""
        if running:
            logging.getLogger("console").debug("Setting RUNNING")

            self.status |= ProcessStatus.RUNNING
            self.status &= ~(ProcessStatus.ENDED | ProcessStatus.NONE)
            self.has_run = True
        else:
            logging.getLogger("console").debug("Unsetting RUNNING")

            self.status &= ~(ProcessStatus.RUNNING | ProcessStatus.NONE)

    @property
    def ended(self) -> bool:
        """Returns True if the process has ended."""
        return ProcessStatus.ENDED in self.status

    @ended.setter
    def ended(self, ended: bool):
        """Set the ended flag."""
        if ended:
            logging.getLogger("console").debug("Setting ENDED")

            self.status |= ProcessStatus.ENDED
            self.status &= ~(ProcessStatus.RUNNING | ProcessStatus.NONE)
            self.has_run = True
        else:
            logging.getLogger("console").debug("Unsetting ENDED")

            self.status &= ~(ProcessStatus.ENDED | ProcessStatus.NONE)

    @property
    def success(self) -> bool:
        """Returns True if the process is successful."""
        return ProcessStatus.SUCCESS in self.status

    @success.setter
    def success(self, success: bool):
        """Set the success flag and removes the error flag."""
        if success:
            logging.getLogger("console").debug("Setting SUCCESS")

            self.status |= ProcessStatus.SUCCESS
            self.status &= ~(ProcessStatus.ERROR | ProcessStatus.NONE)
        else:
            logging.getLogger("console").debug("Unsetting SUCCESS")

            self.status &= ~(ProcessStatus.SUCCESS | ProcessStatus.NONE)

    @property
    def errored(self) -> bool:
        """Returns True if the process has errored.."""
        return ProcessStatus.ERROR in self.status

    @errored.setter
    def errored(self, errored: bool):
        """Set the error flag and removes the success flag."""
        if errored:
            logging.getLogger("console").debug("Setting ERROR")

            self.status |= ProcessStatus.ERROR
            self.status &= ~(ProcessStatus.SUCCESS | ProcessStatus.NONE)
        else:
            logging.getLogger("console").debug("Unsetting ERROR")

            self.status &= ~(ProcessStatus.ERROR | ProcessStatus.NONE)



from __future__ import annotations

import argparse

from tbc_video_export.common.enums import FieldOrder
from tbc_video_export.opts import opt_types, opt_validators


def add_ffmpeg_opts(parent: argparse.ArgumentParser) -> None:
    """Add FFmpeg opts to the parent arg parser."""
    # ffmpeg arguments
    ffmpeg_opts = parent.add_argument_group("ffmpeg")

    ffmpeg_opts.add_argument(
        "--ffmpeg-threads",
        type=int,
        metavar="int",
        help=(
            "Specify the number of FFmpeg threads.\n"
            "  - This overrides --threads, and setting to 0 uses process defaults."
            "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--audio-track",
        dest="audio_track",
        action="append",
        default=[],
        type=opt_validators.validate_audio_track_opts,
        metavar="file_name",
        help="Audio track to mux.\n  - You can use this option multiple times.\n\n",
    )

    ffmpeg_opts.add_argument(
        "--audio-track-advanced",
        dest="audio_track",
        action="append",
        default=[],
        type=opt_validators.validate_audio_track_advanced_opts,
        metavar=(
            "["
            "file_name, "
            "title, "
            "language, "
            "rate, "
            "sample_format, "
            "channels, "
            "channel_layout, "
            "offset"
            "]"
        ),
        help=argparse.SUPPRESS,
        # help=(
        #    "Audio track to mux (advanced).\n"
        #    "  - You can use this option multiple times.\n"
        #    "  - Only file_name is required. None must be used when skipping an arg."
        #    "\n\n"
        #    "Examples:\n"
        #    '  \'["/path/file.flac", "HiFi", "eng", 192000]\'\n'
        #    '  \'["/path/file.flac", "Linear", "eng", None, None, None, None, 0.15]\'\n'  # noqa: E501
        #    '  \'["/path/file.pcm", "Analog Audio", None, 44100, "s16le", 2]\'\n'
        #    '  \'["/path/file.dts", "PCM Surround", "eng", 44100, "s16le", 6, "5.1"]\''
        #    "\n\n"
        # ),
    )

    ffmpeg_opts.add_argument(
        "--metadata",
        nargs=2,
        default=[],
        action="append",
        metavar=("key", "value"),
        help=(
            "Add metadata to output the file.\n"
            "  - You can use this option multiple times."
            "\n\n"
            "Example:\n"
            "  --metadata Title foo --metadata Year 2024"
            "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--metadata-file",
        type=opt_validators.valiate_metadata_file_exists,
        default=[],
        action="append",
        metavar="filename",
        help=(
            "Add metadata to the output of the file using ffmetadata files.\n"
            "  - You can use this option multiple times.\n"
            "  - See https://ffmpeg.org/ffmpeg-formats.html#Metadata-1 for details.\n"
            "  - When using --export-metadata the generated file is also used.\n"
            "  - Files defined here are used before any generated metadata and take\n"
            "    priority."
            "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--field-order",
        type=opt_types.TypeFieldOrder(parent),
        choices=list(FieldOrder),
        default=FieldOrder.TFF,
        metavar="order",
        help=(
            "Set a field order. (default: tff)"
            "\n\n"
            "Available formats:\n  "
            + "\n  ".join(f"{e.name.lower()!s:<5s} {e.value}" for e in FieldOrder)
            + "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--force-anamorphic",
        action="store_true",
        default=False,
        help=(
            "Force widescreen aspect ratio.\n"
            "  - This is not allowed when letterbox is used."
            "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--force-black-level",
        type=opt_validators.validate_black_levels_opts,
        metavar="R[,G,B]",
        default=None,
        help=(
            "Force black levels using the colorlevels filter.\n"
            "  - Use comma-separated numbers.\n"
            "  - If a single number is provided, it is used for all 3."
            "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--hwaccel-device",
        type=str,
        help=(
            "Hardware acceleration device.\n"
            "  - Specify a hardware device when a supported video profile is selected."
            "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--no-attach-json",
        action="store_true",
        default=False,
        help="Disable embedding the TBC json in the video file. (default: no).\n\n",
    )

    ffmpeg_opts.add_argument(
        "--thread-queue-size",
        type=int,
        default=1024,
        metavar="int",
        help=(
            "Set the thread queue size for FFmpeg. (default: 1024)\n"
            "  - Reduce this if you are having out-of-memory issues as a higher value\n"
            "    will allow the decoder to consume more memory."
            "\n\n"
        ),
    )

    ffmpeg_opts.add_argument(
        "--checksum",
        action="store_true",
        default=False,
        help=(
            "Enable SHA256 checksumming on the output streams. (default: no)\n"
            "  - This will create a .sha256 file next to your output file.\n"
            "  - This may reduce export FPS slightly.\n"
            "  - FFmpeg must be used to verify checksums."
            "\n\n"
        ),
    )
